classdef subject_blade_legs < handle
    properties
        L
        Lp % pelvis
        Lto % torso
        Lt % thigh
        Ls % shank
        Lf % foot
        
        c % relative centers of mass
        cto
        ct
        cs
        cf
        
        mmat % inertia matrix
        m % relative masses
        mp
        mto
        mt
        ms
        mf
        
        s % scale factors
        sp1 % frontal: height from width
        sp2 % sagittal depth from frontal width
        sto % sagittal depth from height
        st
        ss
        sf1 % length
        sf2 % width
        sf3 % height
        
        I
        Ip
        Ito
        It
        Is
        If
    end
    
    methods 

        function obj = subject_blade_legs(L,varargin)
            numvarargs = length(varargin);
            if numvarargs > 8
                error('subject_blade_legs:TooManyInputs', ...
                'requires at most 4 optional inputs');
            end
            if mod(numvarargs,2) == 1
                error('subject_blade_legs:Invalid size of optional inputs');
            end
            
            ind = find(strcmp(varargin, 'c'));
            if ~isempty(ind) && ind < numvarargs
                obj.c = varargin{ind+1};
                obj.cto = obj.c(1);
                obj.ct = obj.c(2);
                obj.cs = obj.c(3);
                obj.cf = obj.c(4);
            end
            ind = find(strcmp(varargin, 'm'));
            if ~isempty(ind) && ind < numvarargs
                obj.m = varargin{ind+1};
                obj.mp = obj.m(1);
                obj.mto = obj.m(2);
                obj.mt = obj.m(3);
                obj.ms = obj.m(4);
                obj.mf = obj.m(5);
            end
            ind = find(strcmp(varargin, 's'));
            if ~isempty(ind) && ind < numvarargs
                obj.s = varargin{ind+1};
                obj.sp1 = obj.s(1);
                obj.sp2 = obj.s(2);
                obj.sto = obj.s(3);
                obj.st = obj.s(4);
                obj.ss = obj.s(5);
                obj.sf1 = obj.s(6);
                obj.sf2 = obj.s(7);
                obj.sf3 = obj.s(8);
            end
            ind = find(strcmp(varargin, 'I'));
            if ~isempty(ind) && ind < numvarargs
                obj.I = varargin{ind+1};
                obj.Ip = obj.I(:,1);
                obj.Ito = obj.I(:,2);
                obj.It = obj.I(:,3);
                obj.Is = obj.I(:,4);
                obj.If = obj.I(:,5);
            end
            if size(L,1) > size(L,2)
                L = L';
            end
            obj.L = L;
            obj.Lp = L(1);
            obj.Lto = L(2);
            obj.Lt = L(3);
            obj.Ls = L(4);
            obj.Lf = L(5);
            
            if isempty(obj.c)
                % Dempster body parameters
                obj.c = [0.626, 0.433, 0.433, 0.5].*obj.L(2:end);
                obj.cto = obj.c(1);
                obj.ct = obj.c(2);
                obj.cs = obj.c(3);
                obj.cf = obj.c(4);
            end
            if isempty(obj.m)
                % Dempster body parameters
                obj.m = [0.142, 0.536, 0.1, 0.0465, 0.0145];
                obj.mp = obj.m(1);
                obj.mto = obj.m(2);
                obj.mt = obj.m(3);
                obj.ms = obj.m(4);
                obj.mf = obj.m(5);
            end
            if isempty(obj.I)
                if isempty(obj.s)
                    % average scale parameters from body surface scans
                    obj.s = [0.2, 0.236, 0.673, 0.376, 0.296, 1, 0.362, 0.257];
                    obj.sp1 = obj.s(1);
                    obj.sp2 = obj.s(2);
                    obj.sto = obj.s(3);
                    obj.st = obj.s(4);
                    obj.ss = obj.s(5);
                    obj.sf1 = obj.s(6);
                    obj.sf2 = obj.s(7);
                    obj.sf3 = obj.s(8);
                end
                getInertia(obj);
            end
        end
        
        function getInertia(obj)
            % circular cylinders: upper arm, forearm, thigh, shank
            rt = obj.st*obj.Lt/2;
            obj.It = obj.mt*[1/12*(3*rt^2+obj.Lt^2); 1/2*rt^2; 1/12*(3*rt^2+obj.Lt^2)];
            rs = obj.ss*obj.Ls/2;
            obj.Is = obj.ms*[1/12*(3*rs^2+obj.Ls^2); 1/2*rs^2; 1/12*(3*rs^2+obj.Ls^2)];
            
            % elliptical cylinder: pelvis, torso
            obj.Ip = zeros(3,1);
            Lpv = obj.Lp*obj.sp1;
            Lpd = obj.Lp*obj.sp2;
            obj.Ip(2) = obj.mp*(obj.Lp/2)^2/4+obj.mp*Lpv^2/12; % x in "Kinetics of Human Motion" by Zatsiorsky
            obj.Ip(3) = obj.mp*(Lpd/2)^2/4+obj.mp*Lpv^2/12; % y
            obj.Ip(1) = obj.mp*((obj.Lp/2)^2+(Lpd/2)^2)/4; % z
            
            % local joint coord: x = 1 = proximal-distal, y = 2 = frontal-back, z = 3 =
            % medial-lateral
            obj.Ito = zeros(3,1);
            Ltod = (obj.Lto)*obj.sto;
            obj.Ito(3) = obj.mp*(obj.Lp/2)^2/4+obj.mp*obj.Lto^2/12;
            obj.Ito(2) = obj.mp*(Ltod/2)^2/4+obj.mp*obj.Lto^2/12;
            obj.Ito(1) = obj.mp*((obj.Lp/2)^2+(Ltod/2)^2)/4;
            
            % semi ellipsoid: foot
            % local joint coord: x = 1 = frontal, y = 2 = sagittal, z = 3 = vertical
            rf1 = obj.sf1*obj.Lf; % length
            rf2 = obj.sf2*obj.Lf/2; % width
            rf3 = obj.sf3*obj.Lf/2; % height
            obj.If = obj.mf*[1/5*(rf3^2+rf2^2); 1/5*(rf3^2+rf1^2)-(3/8*rf1)^2; 1/5*(rf2^2+rf1^2)-(3/8*rf1)^2];
            
            obj.I = [obj.Ip; obj.Ito; obj.It; obj.Is; obj.If; obj.It; obj.Is; obj.If];
            
            updateMassMat(obj);
        end
        
        function updateMassMat(obj)
            obj.mmat = diag([obj.m(1)*ones(3,1);obj.m(2)*ones(3,1);obj.m(3)*ones(3,1);obj.m(4)*ones(3,1);obj.m(5)*ones(3,1);obj.m(3)*ones(3,1);...
                obj.m(4)*ones(3,1);obj.m(5)*ones(3,1);obj.I]);
        end
        
    end
end
